/***
 * Juliac
 * Copyright (C) 2007-2012 Inria, Univ. Lille 1
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Lionel Seinturier
 */

package org.objectweb.fractal.juliac.opt.oo;

import java.io.IOException;
import java.util.List;

import org.objectweb.fractal.api.type.ComponentType;
import org.objectweb.fractal.api.type.InterfaceType;
import org.objectweb.fractal.julia.loader.BasicLoader;
import org.objectweb.fractal.juliac.Utils;
import org.objectweb.fractal.juliac.api.SpoonSupportItf;
import org.objectweb.fractal.juliac.conf.JuliacConfig;
import org.objectweb.fractal.juliac.desc.ControllerDesc;
import org.objectweb.fractal.juliac.desc.MembraneDesc;
import org.objectweb.fractal.juliac.desc.SimpleMembraneDesc;
import org.objectweb.fractal.juliac.opt.FCSourceCodeGenerator;
import org.objectweb.fractal.juliac.proxy.ProxyClassGeneratorItf;

/**
 * This class generates the source code associated to Fractal components.
 * The membrane implementation and the initializer implementation are generated.
 * 
 * The content, the interceptors and the controllers are kept in separate
 * classes and the controllers are implemented with objects.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 */
public class FCOOCtrlSourceCodeGenerator
extends FCSourceCodeGenerator<List<ControllerDesc>> {
    
    @Override
    protected void postInitJuliaLoaderModule() throws IOException {
        jloader.loadConfigFile(BasicLoader.DEFAULT_CONFIGURATION);
        super.postInitJuliaLoaderModule();
    }

    
    // -----------------------------------------------------------------------
    // Implementation of the FCSourceCodeGeneratorItf interface
    // -----------------------------------------------------------------------

    /**
     * Generate the implementation of the membrane associated to a Fractal
     * component.
     * 
     * @param ct                the component type
     * @param ctrldesc          the component controller descriptor
     * @param contentClassName  the name of the component content class
     * @return                  data about the membrane implementation
     */
    public MembraneDesc<List<ControllerDesc>> generateMembraneImpl(
        ComponentType ct, String ctrldesc, String contentClassName )
    throws IOException {
        
    	/*
         * Generate business and control interface implementations.
         */
        InterfaceType[] its = ct.getFcInterfaceTypes();
        generateInterfaceImpl(its,ctrldesc);
        InterfaceType[] membraneits = mloader.getMembraneType(ct,ctrldesc);
        generateInterfaceImpl(membraneits,ctrldesc);
        
        /*
         * Generate controller implementations.
         */
        List<ControllerDesc> ctrlDescs = mloader.getCtrlImplLayers(ct,ctrldesc);        
        for (ControllerDesc ctrlDesc : ctrlDescs) {
            
            List<String> ctrlImplLayers = ctrlDesc.getLayerImplNames();
            String ctrlImplClassName = ctrlDesc.getImplName();
            
            if( ctrlImplClassName.startsWith("(") ) {
                /*
                 * Skip CloneableAttributeController implementations.
                 * The controller implementation class name is in this case
                 * something like:
                 * 
                 * ((attribute-controller org.objectweb.fractal.julia.conform.components.CAttributes))
                 */
                continue;
            }
            
            if( ctrlImplLayers == null ) {
                /*
                 * Skip the case where the controller is a single class
                 * (not obtained by mixing layers).
                 */
                continue;
            }
            
            if( ctrlImplClassName.length() == 0 ) {
                /*
                 * For controllers implemented with Scala, the membrane
                 * description is typically of the form:
                 * 
                 * 		@Controller(
                 * 			name="component",
                 * 			mixins=ComponentControllerImpl.class)
                 * 		protected Component comp;
                 * 
                 * with an omitted implName and only one layer.
                 */
            	ctrlImplClassName = ctrlDesc.getImplNameOrUniqueLayerImplName();
            	ctrlDesc.setImplName(ctrlImplClassName);
            	continue;
            }
            
            ctrlImplClassName =
            	Utils.getJuliacGeneratedStrongTypeName(ctrlImplClassName);
            ctrlImplClassName = jc.getRootedClassName(ctrlImplClassName);
            ctrlDesc.setImplName(ctrlImplClassName);
            
            // Avoid duplicate generation
            if( ! jc.hasBeenGenerated(ctrlImplClassName) ) {
            	JuliacConfig jconf = jc.getJuliacConfig();
            	SpoonSupportItf spoon =
            		jconf.lookupUnique(SpoonSupportItf.class);
                spoon.mixAndGenerate(ctrlImplClassName,ctrlImplLayers);
            }
        }
        
        /*
         * Prepare a membrane descriptor.
         */
        MembraneDesc<List<ControllerDesc>> md =
            new SimpleMembraneDesc(jc,null,ctrldesc,membraneits,ctrlDescs);
        ProxyClassGeneratorItf pcg =
        	getInterceptorClassGenerator(ct,ctrldesc);
        md.setInterceptorClassGenerator(pcg);

        /*
         * Generate interceptor implementations.
         * Interceptors must be generated after controllers as some interceptors
         * (e.g. the lifecycle interceptor) use controller implementation
         * class names.
         */
        generateInterceptorImpl(its,md);
        generateInterceptorImpl(membraneits,md);
        
        /*
         * Generate the membrane initializer class.
         */
        InitializerOOCtrlClassGenerator icg = getInitializerClassGenerator();
        icg.init(jc,this,md,ct,contentClassName);
        MembraneInitializerOOCtrlClassGenerator micg =
            icg.getMembraneInitializerClassGenerator();
        jc.generateSourceCode(micg);
        
        return md;
    }

    /**
     * Return the class generator for component initializers. An initializer is
     * a factory and provides a <code>newFcInstance</code> method for
     * instantiating components
     */
    public InitializerOOCtrlClassGenerator getInitializerClassGenerator() {
        return new InitializerOOCtrlClassGenerator();
    }
}
