/***
 * Juliac
 * Copyright (C) 2007-2012 Inria, Univ. Lille 1
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Lionel Seinturier
 * Contributor: Romain Rouvoy
 */

package org.objectweb.fractal.juliac.opt.oo;

import java.util.List;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.Interface;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.type.InterfaceType;
import org.objectweb.fractal.julia.InitializationContext;
import org.objectweb.fractal.julia.factory.ChainedInstantiationException;
import org.objectweb.fractal.juliac.Utils;
import org.objectweb.fractal.juliac.api.ClassGeneratorItf;
import org.objectweb.fractal.juliac.api.JuliacRuntimeException;
import org.objectweb.fractal.juliac.desc.ControllerDesc;
import org.objectweb.fractal.juliac.desc.MembraneDesc;
import org.objectweb.fractal.juliac.opt.InitializerClassGenerator;
import org.objectweb.fractal.juliac.proxy.AttributeControllerClassGenerator;
import org.objectweb.fractal.juliac.visit.BlockSourceCodeVisitor;

/**
 * Initializer generator for Fractal components with an object-oriented control
 * membrane and where the content, the interceptors and the controllers are kept
 * in separate classes.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 * @author Romain Rouvoy <Romain.Rouvoy@univ-lille1.fr>
 */
public class InitializerOOCtrlClassGenerator
extends InitializerClassGenerator<List<ControllerDesc>> {
    
    @Override
    public String getSuperClassName() {
        MembraneInitializerOOCtrlClassGenerator micg =
            getMembraneInitializerClassGenerator();
        String name = micg.getTargetClassName();
        return name;
    }

    @Override
    protected void generateGetFcControllerDescMethod( BlockSourceCodeVisitor mv ) {
        mv.visit  ("      return \"");
        MembraneDesc<?> membraneDesc = getMembraneDesc();
        mv.visit  (membraneDesc.getDescriptor());
        mv.visitln("\";");
    }
    
    @Override
    protected void generateGetFcContentDescMethod( BlockSourceCodeVisitor mv ) {
        mv.visit  ("      return ");
        mv.visit  (Utils.javaifyContentDesc(contentDesc).toString());
        mv.visitln(";");
    }
    
    @Override
    protected void generateNewFcContentMethod( BlockSourceCodeVisitor mv ) {        
        
        MembraneDesc<List<ControllerDesc>> membraneDesc = getMembraneDesc();
        String ctrlDesc = membraneDesc.getDescriptor();
        String contentClassName =
            Utils.getContentClassName(ctrlDesc,contentDesc);

        if( contentClassName == null ) {
            mv.visitln("    return null;");
        }
        else {
            mv.visitln("    try {");
            mv.visit  ("      Object content = new ");
            mv.visit  (contentClassName);
            mv.visitln("();");
            mv.visitln("      return content;");
            mv.visitln("    }");
            mv.visitln("    catch( Throwable t ) {");
            mv.visit  ("      throw new ");
            mv.visit  (ChainedInstantiationException.class.getName());
            mv.visitln("(t,null,\"\");");
            mv.visitln("    }");
        }
    }
    
    @Override
    protected void generateNewFcInstanceMethod( BlockSourceCodeVisitor mv ) {
        mv.visitln("    Object content = newFcContent();");
        mv.visitln("    return newFcInstance(content);");
    }        

    @Override
    protected void generateNewFcInstanceContentMethod( BlockSourceCodeVisitor mv ) {
        
        MembraneDesc<List<ControllerDesc>> membraneDesc = getMembraneDesc();
        InterfaceType[] membraneits = membraneDesc.getCtrlItfTypes();
        boolean isComposite = Utils.isComposite(membraneits);
        
        // Check the content
        generateContentChecks(mv);

        // Invoke the newFcInitializationContext method inherited from the MembraneInitializerClassGenerator
        mv.visitln("    "+InitializationContext.class.getName()+" ic = newFcInitializationContext(content);");        
        
        // Store the content in the initialization context
        generateInitializationContextForContent(mv);

        // Declare local variables which will be used in the next steps
        mv.visitln("    "+Interface.class.getName()+" proxy;");
        mv.visitln("    Object intercept;");
        mv.visitln("    "+Component.class.getName()+" proxyForCompCtrl = ("+Component.class.getName()+") ic.interfaces.get(\"component\");");
        
        // /cloneable-attribute-controller
        boolean attrimplgenerated = false;
        for (int i=0; i < membraneits.length; i++) {            
            String itname = membraneits[i].getFcItfName();
            if( itname.equals("/cloneable-attribute-controller") ) {
            	generateNFCICMControlInterfaceCAC(mv,membraneits[i]);
                attrimplgenerated = true;
            }
        }
        
        // Component type
        InterfaceType[] its = ct.getFcInterfaceTypes();
        InterfaceType[] fullits = new InterfaceType[ membraneits.length + its.length ];
        System.arraycopy(membraneits, 0, fullits, 0, membraneits.length);
        System.arraycopy(its, 0, fullits, membraneits.length, its.length);
        mv.visitln("    ic.type = "+Utils.javaify(fullits)+";");
        
        // Business interfaces
        for (int i = 0; i < its.length; i++) {
            
            String itname = its[i].getFcItfName();
            
            if( its[i].isFcCollectionItf() ) {
                // Julia naming convention for collection interfaces
                itname = "/collection/" + itname;
            }

            // Interceptors
            if( ! generateInterceptorCreation(mv,its[i],"content") ) {
                mv.visit("    intercept = ");
                mv.visit( its[i].isFcClientItf() ? "null" : "content" );
                mv.visitln(";");
            }

            // External interface
            generateNFICMExternalInterface(mv,its[i],attrimplgenerated);
            
            // Internal interface (skip attribute control interfaces and primitive components)
            if( ! itname.equals("attribute-controller") && isComposite ) {
                InterfaceType intit = Utils.newSymetricInterfaceType(its[i]);
                ClassGeneratorItf itfscg = fcscg.getInterfaceClassGenerator(intit,membraneDesc);
                String fcitfClassname = itfscg.getTargetClassName();
                String itstring = Utils.javaify(intit).toString();
                mv.visit  ("    proxy = ");
                mv.visit  ("new "+fcitfClassname+"(");
                mv.visit  ("proxyForCompCtrl,\""+itname+"\","+itstring);
                mv.visitln(",true,intercept);");  // interceptor or content or null
                mv.visitln("    ic.internalInterfaces.put(\""+itname+"\",proxy);");
            }
        }
        
        // Initialize controllers
        mv.visitln("    initFcController(ic);");
        
        // Return the fcinterface for the component controller
        mv.visitln("    return proxyForCompCtrl;");
    }
    
    /**
     * CAC is a shortcut for /cloneable-attribute-controller.
     * 
     * @since 2.3
     */
    private void generateNFCICMControlInterfaceCAC(
        BlockSourceCodeVisitor mv, InterfaceType membraneit ) {

        String itname = membraneit.getFcItfName();
        ClassGeneratorItf itfscg =
            fcscg.getInterfaceClassGenerator(membraneit,membraneDesc);
        String fcitfClassname = itfscg.getTargetClassName();
        String itstring = Utils.javaify(membraneit).toString();
        
        InterfaceType attrit = null;
        try {
            attrit = ct.getFcInterfaceType("attribute-controller");
        }
        catch (NoSuchInterfaceException e) {
            throw new JuliacRuntimeException(e);
        }
        
        ClassGeneratorItf attrscg =
            new AttributeControllerClassGenerator(jc,attrit,null,false);
        String s = attrscg.getTargetClassName();
        
        mv.visitIns("Object attrimpl = new ",s,"()");
        mv.visitIns(
    		"proxy = ","new ",fcitfClassname,"(","proxyForCompCtrl,\""+itname+
    		"\",",itstring,",false,attrimpl)");
        mv.visitIns("ic.interfaces.put(\""+itname+"\",proxy)");
    }
    
    protected void generateNFICMExternalInterface(
        BlockSourceCodeVisitor mv, InterfaceType it,
        boolean attrimplgenerated ) {
        
        String itname = it.getFcItfName();
        
        if( it.isFcCollectionItf() ) {
            // Julia naming convention for collection interfaces
            itname = "/collection/" + itname;
        }

        // External interface
        MembraneDesc<?> membraneDesc = getMembraneDesc();
        ClassGeneratorItf itfscg =
            fcscg.getInterfaceClassGenerator(it,membraneDesc);
        String fcitfClassname = itfscg.getTargetClassName();
        String itstring = Utils.javaify(it).toString();
        mv.visit("    proxy = ");
        mv.visit("new "+fcitfClassname+"(");
        mv.visit("proxyForCompCtrl,\""+itname+"\","+itstring+",false,");
        if( itname.equals("attribute-controller") && attrimplgenerated ) {
            mv.visit("attrimpl");
        }
        else {
            mv.visit("intercept");  // interceptor or content or null
        }
        mv.visitln(");");
        mv.visitln("    ic.interfaces.put(\""+itname+"\",proxy);");
    }

    
    // ----------------------------------------------------------------------
    // Membrane initializer class generator
    // ----------------------------------------------------------------------
    
    protected MembraneInitializerOOCtrlClassGenerator
    getMembraneInitializerClassGenerator() {
        MembraneInitializerOOCtrlClassGenerator micg =
            new MembraneInitializerOOCtrlClassGenerator(
                jc,fcscg,this,membraneDesc,ct,contentDesc);
        return micg;
    }
}
