/***
 * Juliac
 * Copyright (C) 2009-2012 Inria, Univ. Lille 1
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Lionel Seinturier
 * Contributor: Romain Rouvoy
 */

package org.objectweb.fractal.juliac.opt.oo;

import java.lang.reflect.Modifier;
import java.util.HashMap;
import java.util.List;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.Interface;
import org.objectweb.fractal.api.factory.InstantiationException;
import org.objectweb.fractal.api.type.ComponentType;
import org.objectweb.fractal.api.type.InterfaceType;
import org.objectweb.fractal.julia.InitializationContext;
import org.objectweb.fractal.julia.factory.ChainedInstantiationException;
import org.objectweb.fractal.julia.loader.Initializable;
import org.objectweb.fractal.julia.loader.Tree;
import org.objectweb.fractal.julia.type.BasicInterfaceType;
import org.objectweb.fractal.juliac.Juliac;
import org.objectweb.fractal.juliac.Utils;
import org.objectweb.fractal.juliac.api.ClassGeneratorItf;
import org.objectweb.fractal.juliac.api.FCSourceCodeGeneratorItf;
import org.objectweb.fractal.juliac.conf.Constants;
import org.objectweb.fractal.juliac.desc.ControllerDesc;
import org.objectweb.fractal.juliac.desc.MembraneDesc;
import org.objectweb.fractal.juliac.opt.ClassGenerator;
import org.objectweb.fractal.juliac.ucf.UnifiedClass;
import org.objectweb.fractal.juliac.visit.BlockSourceCodeVisitor;
import org.objectweb.fractal.juliac.visit.ClassSourceCodeVisitor;

/**
 * Generator for object-oriented control membrane initializer. An membrane
 * initializer provides the code which is common to all component initializers
 * for a given membrane descriptor.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 * @author Romain Rouvoy <Romain.Rouvoy@univ-lille1.fr>
 * @since 2.2.2
 */
public class MembraneInitializerOOCtrlClassGenerator extends ClassGenerator {

	private Juliac jc;
    protected FCSourceCodeGeneratorItf<List<ControllerDesc>> fcscg;
    protected MembraneDesc<List<ControllerDesc>> membraneDesc;
    protected InitializerOOCtrlClassGenerator icg;
    protected ComponentType ct;
    protected Object contentDesc;
    
    public MembraneInitializerOOCtrlClassGenerator(
        
    	Juliac jc, FCSourceCodeGeneratorItf<List<ControllerDesc>> fcscg,
        InitializerOOCtrlClassGenerator icg,
        MembraneDesc<List<ControllerDesc>> membraneDesc, ComponentType ct,
        Object contentDesc ) {
        
    	this.jc = jc;
        this.fcscg = fcscg;
        this.icg = icg;
        this.membraneDesc = membraneDesc;
        this.ct = ct;
        this.contentDesc = contentDesc;
    }
    
    public String getTargetClassName() {
        String ctrlDesc = membraneDesc.getDescriptor();
        String strong = Utils.getJuliacGeneratedStrongTypeName(ctrlDesc);
        String rooted = jc.getRootedClassName(strong);
        return rooted;
    }
    
    @Override
    public String getSuperClassName() {
        return Constants.JULIAC_RUNTIME_MEMBRANE_INITIALIZER;
    }

    @Override
    public void generateMethods( ClassSourceCodeVisitor cv ) {
        generateNewFcInitializationContextMethod(cv);
        generateNewFcControllerInstantiationMethod(cv);
    }
    
    protected void generateNewFcInitializationContextMethod( ClassSourceCodeVisitor cv ) {
        BlockSourceCodeVisitor mv =
            cv.visitMethod(
                Modifier.PUBLIC, null, InitializationContext.class.getName(),
                "newFcInitializationContext", new String[]{"Object content"},
                new String[]{InstantiationException.class.getName()} );
        generateNewFcInitializationContextMethod(mv);
        mv.visitEnd();        
    }

    protected void generateNewFcInitializationContextMethod( BlockSourceCodeVisitor mv ) {
        
        InterfaceType[] membraneits = membraneDesc.getCtrlItfTypes();
        boolean isComposite = Utils.isComposite(membraneits);
        
        // Prepare an InitializationContext
        mv.visitln("    "+InitializationContext.class.getName()+" ic = initFcInitializationContext();");
        if(isComposite) {
            mv.visitln("    ic.internalInterfaces = new "+HashMap.class.getName()+"();");
        }
        
        // Instantiate controllers
        mv.visitln("    "+Component.class.getName()+" compctrl = newFcControllerInstantiation(ic);");

        InterfaceType compctrlit =
            new BasicInterfaceType(
                "component",Component.class.getName(),false,false,false);

        // Interceptor
        mv.visitln("    Object intercept = null;");
        if( ! icg.generateInterceptorCreation(mv,compctrlit,"compctrl") ) {
            mv.visitln("    intercept = compctrl;");
        }

        // Component controller
        ClassGeneratorItf itfscg =
            fcscg.getInterfaceClassGenerator(compctrlit,membraneDesc);
        String fcitfClassname = itfscg.getTargetClassName();
        generateNCICMProxyCreation(
    		mv,true,fcitfClassname,"compctrl","component",compctrlit);
        mv.visit  ("    "+Component.class.getName()+" proxyForCompCtrl = ");
        mv.visitln("("+Component.class.getName()+") proxy;");
        
        // Controllers other than the component controller
        mv.visitln("    Object ctrl = null;");
        for (int i=0; i < membraneits.length; i++) {
        	
        	InterfaceType membraneit = membraneits[i];
        	String itname = membraneit.getFcItfName();
        	
            // Skip component and /cloneable-attribute-controller
            if( itname.equals("component") ||
            	itname.equals("/cloneable-attribute-controller") ) {
            	continue;
            }
            
            // Control external interface
            mv.visitIns("ctrl = getFcController(ic,content,",membraneit.getFcItfSignature(),".class)");

            // Interceptors
            if( ! icg.generateInterceptorCreation(mv,membraneit,"ctrl") ) {
                mv.visitln("    intercept = ctrl;");
            }

            // Proxy for the interface
            itfscg = fcscg.getInterfaceClassGenerator(membraneit,membraneDesc);
            fcitfClassname = itfscg.getTargetClassName();
            generateNCICMProxyCreation(
        		mv,false,fcitfClassname,"proxyForCompCtrl",itname,membraneit);
            
            // Control internal interface
            if( itname.equals("factory") && isComposite ) {
                InterfaceType intit = Utils.newSymetricInterfaceType(membraneit);
                itfscg = fcscg.getInterfaceClassGenerator(intit,membraneDesc);
                fcitfClassname = itfscg.getTargetClassName();
                String itstring = Utils.javaify(intit).toString();
                mv.visit  ("    proxy = ");
                mv.visit  ("new "+fcitfClassname+"(");
                mv.visit  ("proxyForCompCtrl,\""+itname+"\","+itstring);
                mv.visitln(",true,intercept);");
                mv.visitln("    ic.internalInterfaces.put(\""+itname+"\",proxy);");
            }
        }
        
        mv.visitln("    return ic;");
    }    

    protected void generateNewFcControllerInstantiationMethod( ClassSourceCodeVisitor cv ) {
        BlockSourceCodeVisitor mv =
            cv.visitMethod(
                Modifier.PUBLIC, null, Component.class.getName(),
                "newFcControllerInstantiation",
                new String[]{InitializationContext.class.getName()+" ic"},
                new String[]{InstantiationException.class.getName()} );
        generateNewFcControllerInstantiationMethod(mv);
        mv.visitEnd();        
    }

    protected void generateNewFcControllerInstantiationMethod( BlockSourceCodeVisitor mv ) {
    	
        List<ControllerDesc> ctrlDescs = membraneDesc.getCtrlDescs();
        UnifiedClass ucinitializable = jc.create(Initializable.class.getName());
        UnifiedClass uccomponent = jc.create(Component.class.getName());
        
        mv.visitln("    Object ctrl = null;");

        for (ControllerDesc ctrlDesc : ctrlDescs) {
            String ctrlImplName = ctrlDesc.getImplName();
            if( ctrlImplName.charAt(0) == '(' ) {
                /*
                 * Skip CloneableAttributeController implementations.
                 * The controller implementation class name is in this case
                 * something like:
                 * 
                 * ((attribute-controller org.objectweb.fractal.julia.conform.components.CAttributes))
                 */
                continue;
            }
            mv.visitln("    ic.controllers.add(ctrl = new "+ctrlImplName+"());");
            UnifiedClass uc = jc.create(ctrlImplName);
            if( ucinitializable.isAssignableFrom(uc) ) {
                Tree tree = ctrlDesc.getTree();
                if( tree != null ) {
                    mv.visitln("    "+Tree.class.getName()+" tree = "+Utils.javaify(tree)+';');
                    mv.visitln("    try {");
                    mv.visitln("      (("+Initializable.class.getName()+")ctrl).initialize(tree);");
                    mv.visitln("    }");
                    mv.visitln("    catch( Exception e ) {");
                    mv.visitln("      throw new "+ChainedInstantiationException.class.getName()+"(e,null,\"\");");
                    mv.visitln("    }");
                }
            }
            if( uccomponent.isAssignableFrom(uc) ) {
                mv.visitln("    "+Component.class.getName()+" compctrl = ("+Component.class.getName()+")ctrl;");
            }
        }

        mv.visitln("    return compctrl;");
    }
    
    protected void generateNCICMProxyCreation(
		BlockSourceCodeVisitor mv, boolean typedef, String fcitfClassname,
		String delegate, String itname, InterfaceType it ) {
        
        String itstring = Utils.javaify(it).toString();
    	mv.visitIns(
			typedef ? Interface.class.getName() : "","proxy = new",
			fcitfClassname,"(",delegate,",\""+itname+"\",",itstring,
			",false,intercept)");
        mv.visitIns("ic.interfaces.put(\""+itname+"\",proxy)");    	
    }
}
